@@ -10,15 +10,16 @@ from typing import TYPE_CHECKING
 import xcffib
 import xcffib.xproto
 from xcffib.wrappers import GContextID, PixmapID
-from xcffib.xproto import EventMask, SetMode, StackMode
+from xcffib.xproto import EventMask, SetMode
 
-from libqtile import hook, utils
+from libqtile import bar, hook, utils
 from libqtile.backend import base
 from libqtile.backend.base import FloatStates
 from libqtile.backend.x11 import xcbq
 from libqtile.backend.x11.drawer import Drawer
-from libqtile.command.base import CommandError
+from libqtile.command.base import CommandError, expose_command
 from libqtile.log_utils import logger
+from libqtile.scratchpad import ScratchPad
 
 if TYPE_CHECKING:
     from libqtile.command.base import ItemT
@@ -412,14 +413,7 @@ class XWindow:
         return self.conn.conn.core.GetWindowAttributes(self.wid).reply()
 
     def query_tree(self):
-        q = self.conn.conn.core.QueryTree(self.wid).reply()
-        root = None
-        parent = None
-        if q.root:
-            root = XWindow(self.conn, q.root)
-        if q.parent:
-            parent = XWindow(self.conn, q.parent)
-        return root, parent, [XWindow(self.conn, i) for i in q.children]
+        return self.conn.conn.core.QueryTree(self.wid).reply().children
 
     def paint_borders(self, depth, colors, borderwidth, width, height):
         """
@@ -507,6 +501,13 @@ class _Window:
         self._float_width: int = self._width
         self._float_height: int = self._height
 
+        # We use `previous_layer` to see if a window has moved up or down a "layer"
+        # The layers are defined in the spec:
+        # https://specifications.freedesktop.org/wm-spec/1.3/ar01s07.html#STACKINGORDER
+        # We assume a window starts off in the layer for "normal" windows, i.e. ones that
+        # don't match the requirements to be in any of the other layers.
+        self.previous_layer = (False, False, True, False, False, False)
+
         self.bordercolor = None
         self.state = NormalState
         self._float_state = FloatStates.NOT_FLOATING
@@ -649,13 +650,23 @@ class _Window:
     def update_state(self):
         triggered = ["urgent"]
 
-        if self.qtile.config.auto_fullscreen:
-            triggered.append("fullscreen")
-
         state = self.window.get_net_wm_state()
 
+        if self.qtile.config.auto_fullscreen:
+            triggered.append("fullscreen")
+        # This might seem a bit weird but it's to workaround a bug in chromium based clients not properly redrawing
+        # The bug is described in https://github.com/qtile/qtile/issues/4176
+        # This only happens when auto fullscreen is set to false because we then do not obey the disable fullscreen state
+        # So here we simply re-place the window at the coordinates which will magically solve the issue
+        # This only seems to be an issue with unfullscreening, thus we check if we're fullscreen and the window wants to unfullscreen
+        elif self.fullscreen and "fullscreen" not in state:
+            self._reconfigure_floating(new_float_state=FloatStates.FULLSCREEN)
+
         for s in triggered:
-            setattr(self, s, (s in state))
+            attr = s
+            val = s in state
+            if getattr(self, attr) != val:
+                setattr(self, attr, val)
 
     @property
     def urgent(self):
@@ -668,6 +679,7 @@ class _Window:
         if not val:
             self.hints["urgent"] = False
 
+    @expose_command()
     def info(self):
         if self.group:
             group = self.group.name
@@ -723,6 +735,7 @@ class _Window:
             assert hasattr(self, "window")
             self.window.set_property("_NET_WM_WINDOW_OPACITY", real_opacity)
 
+    @expose_command()
     def kill(self):
         if "WM_DELETE_WINDOW" in self.window.get_wm_protocols():
             data = [
@@ -798,6 +811,7 @@ class _Window:
     def get_pid(self):
         return self.window.get_net_wm_pid()
 
+    @expose_command()
     def place(
         self,
         x,
@@ -893,21 +907,268 @@ class _Window:
         self.width = width
         self.height = height
 
-        kwarg = dict(
-            x=x,
-            y=y,
-            width=width,
-            height=height,
-        )
+        self.window.configure(x=x, y=y, width=width, height=height)
+
         if above:
-            kwarg["stackmode"] = StackMode.Above
+            self.change_layer(up=True)
 
-        self.window.configure(**kwarg)
         self.paint_borders(bordercolor, borderwidth)
 
         if send_notify:
             self.send_configure_notify(x, y, width, height)
 
+    def get_layering_information(self) -> tuple[bool, bool, bool, bool, bool, bool]:
+        """
+        Get layer-related EMWH-flags
+        https://specifications.freedesktop.org/wm-spec/1.3/ar01s07.html#STACKINGORDER
+
+        Copied here:
+
+         To obtain good interoperability between different Desktop Environments,
+         the following layered stacking order is recommended, from the bottom:
+
+            - windows of type _NET_WM_TYPE_DESKTOP
+            - windows having state _NET_WM_STATE_BELOW
+            - windows not belonging in any other layer
+            - windows of type _NET_WM_TYPE_DOCK (unless they have state
+              _NET_WM_TYPE_BELOW) and windows having state _NET_WM_STATE_ABOVE
+            - focused windows having state _NET_WM_STATE_FULLSCREEN
+
+        Windows that are transient for another window should be kept above this window.
+
+        The window manager may choose to put some windows in different stacking positions,
+        for example to allow the user to bring currently a active window to the top and return
+        it back when the window loses focus. To this end, qtile adds an additional layer so that
+        scratchpad windows are placed above all others, always.
+        """
+        state = self.window.get_net_wm_state()
+        _type = self.window.get_wm_type() or ""
+
+        # Check if this window is focused
+        active_window = self.qtile.core._root.get_property(
+            "_NET_ACTIVE_WINDOW", "WINDOW", unpack=int
+        )
+        if active_window and active_window[0] == self.window.wid:
+            focus = True
+        else:
+            focus = False
+
+        desktop = _type == "desktop"
+        below = "_NET_WM_STATE_BELOW" in state
+        dock = _type == "dock"
+        above = "_NET_WM_STATE_ABOVE" in state
+        full = (
+            "fullscreen" in state
+        )  # get_net_wm_state translates this state so we don't use _NET_WM name
+        is_scratchpad = isinstance(self.qtile.groups_map.get(self.group), ScratchPad)
+
+        # sort the flags from bottom to top, True meaning further below than False at each step
+        states = [desktop, below, above or (dock and not below), full and focus, is_scratchpad]
+        other = not any(states)
+        states.insert(2, other)
+
+        # If we're a desktop, this should always be the lowest layer...
+        if desktop:
+            # mypy can't work out that this gives us tuple[bool, bool, bool, bool, bool, bool]...
+            # (True, False, False, False, False, False)
+            return tuple(not i for i in range(6))  # type: ignore
+
+        # ...otherwise, we set to the highest matching layer.
+        # Look for the highest matching level and then set all other levels to False
+        highest = max(i for i, state in enumerate(states) if state)
+
+        # mypy can't work out that this gives us tuple[bool, bool, bool, bool, bool, bool]...
+        return tuple(i == highest for i in range(6))  # type: ignore
+
+    def change_layer(self, up=True, top_bottom=False):
+        """Raise a window above its peers or move it below them, depending on 'up'.
+        Raising a normal window will not lift it above pinned windows etc.
+
+        There are a few important things to take note of when relaying windows:
+        1. If a window has a defined parent, it should not be moved underneath it.
+           In case children are blocking, this could leave an application in an unusable state.
+        2. If a window has children, they should be moved along with it.
+        3. If a window has a defined parent, either move the parent or do nothing at all.
+        4. EMWH-flags follow strict layering rules:
+           https://specifications.freedesktop.org/wm-spec/1.3/ar01s07.html#STACKINGORDER
+        """
+        if len(self.qtile.windows_map) < 2:
+            return
+
+        if self.group is None and not isinstance(self, Static):
+            return
+
+        # Use the window's group or current group if this isn't set (e.g. Static windows)
+        group = self.group or self.qtile.current_group
+
+        parent = self.window.get_wm_transient_for()
+        if parent is not None and not up:
+            return
+
+        layering = self.get_layering_information()
+
+        # Comparison of layer states: -1 if window is now in a lower state group,
+        # 0 if it's in the same group and 1 if it's in a higher group
+        moved = (self.previous_layer > layering) - (layering > self.previous_layer)
+        self.previous_layer = layering
+
+        stack = list(self.qtile.core._root.query_tree())
+        if self.wid not in stack or len(stack) < 2:
+            return
+
+        # Get all windows for the group and add Static windows to ensure these are included
+        # in the stacking
+        group_windows = group.windows.copy()
+        statics = [win for win in self.qtile.windows_map.values() if isinstance(win, Static)]
+        group_windows.extend(statics)
+
+        if group.screen is not None:
+            group_bars = [gap for gap in group.screen.gaps if isinstance(gap, bar.Bar)]
+        else:
+            group_bars = []
+
+        # Get list of windows that are in the stack and managed by qtile
+        # List of tuples (XWindow object, transient_for, layering_information)
+        windows = list(
+            map(
+                lambda w: (
+                    w.window,
+                    w.window.get_wm_transient_for(),
+                    w.get_layering_information(),
+                ),
+                group_windows,
+            )
+        )
+
+        # Sort this list to match stacking order reported by server
+        windows.sort(key=lambda w: stack.index(w[0].wid))
+
+        # Get lists of windows on lower, higher or same "layer" as window
+        lower = [w[0].wid for w in windows if w[1] is None and w[2] > layering]
+        higher = [w[0].wid for w in windows if w[1] is None and w[2] < layering]
+        same = [w[0].wid for w in windows if w[1] is None and w[2] == layering]
+
+        # We now need to identify the new position in the stack
+
+        # If the window has a parent, the window should just be put above it
+        if parent:
+            sibling = parent
+            above = True
+
+        # Now we just check whether the window has changed layer.
+
+        # If we're forcing to top or bottom of current layer...
+        elif top_bottom:
+            # If there are no other windows in the same layer then there's nothing to do
+            if not same:
+                return
+
+            if up:
+                sibling = same[-1]
+                above = True
+            else:
+                sibling = same[0]
+                above = False
+
+        # There are no windows in the desired layer (should never happen) or
+        # we've moved to a new layer and are the only window in that layer
+        elif not same or (len(same) == 1 and moved != 0):
+            # Try to put it above the last window in the lower layers
+            if lower:
+                sibling = lower[-1]
+                above = True
+
+            # Or below the first window in the higher layers
+            elif higher:
+                sibling = higher[0]
+                above = False
+
+            # Don't think we should end up here but, if we do...
+            else:
+                # Put the window above the highest window if we're raising it
+                if up:
+                    sibling = stack[-1]
+                    above = True
+
+                # or below the lowest window if we're lowering the window
+                else:
+                    sibling = stack[0]
+                    above = False
+
+        else:
+            # Window has moved to a lower layer state
+            if moved < 0:
+                if self.kept_below:
+                    sibling = same[0]
+                    above = False
+                else:
+                    sibling = same[-1]
+                    above = True
+
+            # Window is in same layer state
+            elif moved == 0:
+                try:
+                    pos = same.index(self.wid)
+                except ValueError:
+                    pos = len(same) if up else 0
+                if not up:
+                    pos = max(0, pos - 1)
+                else:
+                    pos = min(pos + 1, len(same) - 1)
+                sibling = same[pos]
+                above = up
+
+            # Window is in a higher layer
+            else:
+                if self.kept_above:
+                    sibling = same[-1]
+                    above = True
+                else:
+                    sibling = same[0]
+                    above = False
+
+        # If the sibling is the current window then we just check if any windows in lower/higher layers are
+        # stacked incorrectly and, if so, restack them. However, we don't need to configure stacking for this
+        # window
+        if sibling == self.wid:
+            index = stack.index(self.wid)
+
+            # We need to make sure the bars are included so add them now
+            if group_bars:
+                for group_bar in group_bars:
+                    bar_layer = group_bar.window.get_layering_information()
+                    if bar_layer > layering:
+                        lower.append(group_bar.window.wid)
+                    elif bar_layer < layering:
+                        higher.append(group_bar.window.wid)
+
+                # Sort the list to match the server's stacking order
+                lower.sort(key=lambda wid: stack.index(wid))
+                higher.sort(key=lambda wid: stack.index(wid))
+
+            for wid in [w for w in lower if stack.index(w) > index]:
+                self.qtile.windows_map[wid].window.configure(
+                    stackmode=xcffib.xproto.StackMode.Below, sibling=same[0]
+                )
+
+            # We reverse higher as each window will be placed above the last item in the current layer
+            # this means the last item we stack will be just above the current layer.
+            for wid in [w for w in higher[::-1] if stack.index(w) < index]:
+                self.qtile.windows_map[wid].window.configure(
+                    stackmode=xcffib.xproto.StackMode.Above, sibling=same[-1]
+                )
+
+            return
+
+        # Window needs new stacking info. We tell the server to stack the window
+        # above or below a given "sibling"
+        self.window.configure(
+            stackmode=xcffib.xproto.StackMode.Above if above else xcffib.xproto.StackMode.Below,
+            sibling=sibling,
+        )
+        # TODO: also move our children if we were moved upwards
+        self.qtile.core.update_client_lists()
+
     def paint_borders(self, color, width):
         self.borderwidth = width
         self.bordercolor = color
@@ -982,19 +1243,19 @@ class _Window:
             )
 
             self.window.send_event(e)
+            return True
 
         # we didn't focus this time. but now the window knows if it wants
         # focus, it should SetFocus() itself; we'll get another notification
         # about this.
         return False
 
-    def focus(self, warp: bool) -> None:
+    @expose_command()
+    def focus(self, warp: bool = True) -> None:
+        """Focuses the window."""
         did_focus = self._do_focus()
         if not did_focus:
             return
-        if isinstance(self, base.Internal):
-            # self._do_focus is enough for internal windows
-            return
 
         # now, do all the other WM stuff since the focus actually changed
         if warp and self.qtile.config.cursor_warp:
@@ -1029,17 +1290,19 @@ class _Window:
 
         if self.group:
             self.group.current_window = self
+
+        # Check if we need to restack a previously focused fullscreen window
+        self.qtile.core.check_stacking(self)
+
         hook.fire("client_focus", self)
 
-    def cmd_focus(self, warp: bool = True) -> None:
-        """Focuses the window."""
-        self.focus(warp)
-
-    def cmd_hints(self):
+    @expose_command()
+    def get_hints(self):
         """Returns the X11 hints (WM_HINTS and WM_SIZE_HINTS) for this window."""
         return self.hints
 
-    def cmd_inspect(self):
+    @expose_command()
+    def inspect(self):
         """Tells you more than you ever wanted to know about a window"""
         a = self.window.get_attributes()
         attrs = {
@@ -1092,6 +1355,98 @@ class _Window:
             float_info=float_info,
         )
 
+    @expose_command()
+    def keep_above(self, enable: bool | None = None):
+        if enable is None:
+            self.kept_above = not self.kept_above
+        else:
+            self.kept_above = enable
+
+        self.change_layer(top_bottom=True, up=True)
+
+    @expose_command()
+    def keep_below(self, enable: bool | None = None):
+        if enable is None:
+            self.kept_below = not self.kept_below
+        else:
+            self.kept_below = enable
+
+        self.change_layer(top_bottom=True, up=False)
+
+    @expose_command()
+    def move_up(self, force=False):
+        if self.kept_below and force:
+            self.kept_below = False
+        with self.qtile.core.masked():
+            # Disable masks so that moving windows along the Z axis doesn't trigger
+            # focus change events (i.e. due to `follow_mouse_focus`)
+            self.change_layer()
+
+    @expose_command()
+    def move_down(self, force=False):
+        if self.kept_above and force:
+            self.kept_above = False
+        with self.qtile.core.masked():
+            self.change_layer(up=False)
+
+    @expose_command()
+    def move_to_top(self, force=False):
+        if self.kept_below and force:
+            self.kept_below = False
+        with self.qtile.core.masked():
+            self.change_layer(top_bottom=True)
+
+    @expose_command()
+    def move_to_bottom(self, force=False):
+        if self.kept_above and force:
+            self.kept_above = False
+        with self.qtile.core.masked():
+            self.change_layer(up=False, top_bottom=True)
+
+    @property
+    def kept_above(self):
+        reply = list(self.window.get_property("_NET_WM_STATE", "ATOM", unpack=int))
+        atom = self.qtile.core.conn.atoms["_NET_WM_STATE_ABOVE"]
+        return atom in reply
+
+    @kept_above.setter
+    def kept_above(self, value):
+        reply = list(self.window.get_property("_NET_WM_STATE", "ATOM", unpack=int))
+        atom = self.qtile.core.conn.atoms["_NET_WM_STATE_ABOVE"]
+        if value and atom not in reply:
+            reply.append(atom)
+        elif not value and atom in reply:
+            reply.remove(atom)
+        else:
+            return
+        atom = self.qtile.core.conn.atoms["_NET_WM_STATE_BELOW"]
+        if atom in reply:
+            reply.remove(atom)
+        self.window.set_property("_NET_WM_STATE", reply)
+        self.change_layer()
+
+    @property
+    def kept_below(self):
+        reply = list(self.window.get_property("_NET_WM_STATE", "ATOM", unpack=int))
+        atom = self.qtile.core.conn.atoms["_NET_WM_STATE_BELOW"]
+        return atom in reply
+
+    @kept_below.setter
+    def kept_below(self, value):
+        reply = list(self.window.get_property("_NET_WM_STATE", "ATOM", unpack=int))
+        atom = self.qtile.core.conn.atoms["_NET_WM_STATE_BELOW"]
+        if value and atom not in reply:
+            reply.append(atom)
+        elif not value and atom in reply:
+            reply.remove(atom)
+        else:
+            return
+        atom = self.qtile.core.conn.atoms["_NET_WM_STATE_ABOVE"]
+        if atom in reply:
+            reply.remove(atom)
+        self.window.set_property("_NET_WM_STATE", reply)
+        self.change_layer(up=False)
+
 
 class Internal(_Window, base.Internal):
     """An internal window, that should not be managed by qtile"""
@@ -1114,18 +1469,17 @@ class Internal(_Window, base.Internal):
         win.set_property("QTILE_INTERNAL", 1)
         self._depth = desired_depth
 
-    def create_drawer(self, width: int, height: int) -> base.Drawer:
+    def create_drawer(self, width: int, height: int) -> Drawer:
         """Create a Drawer that draws to this window."""
         return Drawer(self.qtile, self, width, height)
 
+    @expose_command()
     def kill(self):
         if self.window.wid in self.qtile.windows_map:
             # It will be present during config reloads; absent during shutdown as this
             # will follow graceful_shutdown
-            self.qtile.core.conn.conn.core.DestroyWindow(self.window.wid)
-
-    def cmd_kill(self):
-        self.kill()
+            with contextlib.suppress(xcffib.ConnectionException):
+                self.qtile.core.conn.conn.core.DestroyWindow(self.window.wid)
 
     def handle_Expose(self, e):  # noqa: N802
         self.process_window_expose()
@@ -1162,6 +1516,11 @@ class Internal(_Window, base.Internal):
             id=self.window.wid,
         )
 
+    @expose_command()
+    def focus(self, warp: bool = True) -> None:
+        """Focuses the window."""
+        self._do_focus()
+
 
 class Static(_Window, base.Static):
     """An static window, belonging to a screen rather than a group"""
@@ -1259,8 +1618,10 @@ class Static(_Window, base.Static):
         if name == "_NET_WM_STRUT_PARTIAL":
             self.update_strut()
 
-    def cmd_bring_to_front(self):
-        self.window.configure(stackmode=StackMode.Above)
+    @expose_command()
+    def bring_to_front(self):
+        if self.get_wm_type() != "desktop":
+            self.window.configure(stackmode=xcffib.xproto.StackMode.Above)
 
 
 class Window(_Window, base.Window):
@@ -1306,6 +1667,9 @@ class Window(_Window, base.Window):
 
     @floating.setter
     def floating(self, do_float):
+        stack = self.qtile.core._root.query_tree()
+        tiled = [win.window.wid for win in (self.group.tiled_windows if self.group else [])]
+        tiled_stack = [wid for wid in stack if wid in tiled and wid != self.window.wid]
         if do_float and self._float_state == FloatStates.NOT_FLOATING:
             if self.group and self.group.screen:
                 screen = self.group.screen
@@ -1315,9 +1679,21 @@ class Window(_Window, base.Window):
                     self._float_width,
                     self._float_height,
                 )
+
+                # Make sure floating window is placed above tiled windows
+                if tiled_stack and (not self.kept_above or self.qtile.config.floats_kept_above):
+                    stack_list = list(stack)
+                    highest_tile = tiled_stack[-1]
+                    if stack_list.index(self.window.wid) < stack_list.index(highest_tile):
+                        self.window.configure(
+                            stackmode=xcffib.xproto.StackMode.Above, sibling=highest_tile
+                        )
             else:
                 # if we are setting floating early, e.g. from a hook, we don't have a screen yet
                 self._float_state = FloatStates.FLOATING
+            if not self.kept_above and self.qtile.config.floats_kept_above:
+                self.keep_above(enable=True)
+
         elif (not do_float) and self._float_state != FloatStates.NOT_FLOATING:
             self.update_fullscreen_wm_state(False)
             if self._float_state == FloatStates.FLOATING:
@@ -1326,6 +1702,10 @@ class Window(_Window, base.Window):
                 self._float_height = self.height
             self._float_state = FloatStates.NOT_FLOATING
             self.group.mark_floating(self, False)
+            if tiled_stack:
+                self.window.configure(
+                    stackmode=xcffib.xproto.StackMode.Above, sibling=tiled_stack[-1]
+                )
             hook.fire("float_change")
 
     @property
@@ -1336,6 +1716,7 @@ class Window(_Window, base.Window):
             pass
         return False
 
+    @expose_command()
     def toggle_floating(self):
         self.floating = not self.floating
 
@@ -1364,8 +1745,12 @@ class Window(_Window, base.Window):
     @fullscreen.setter
     def fullscreen(self, do_full):
         if do_full:
+            needs_change = self._float_state != FloatStates.FULLSCREEN
             screen = self.group.screen or self.qtile.find_closest_screen(self.x, self.y)
 
+            if self._float_state not in (FloatStates.MAXIMIZED, FloatStates.FULLSCREEN):
+                self._save_geometry()
+
             bw = self.group.floating_layout.fullscreen_border_width
             self._enablefloating(
                 screen.x,
@@ -1374,15 +1759,17 @@ class Window(_Window, base.Window):
                 screen.height - 2 * bw,
                 new_float_state=FloatStates.FULLSCREEN,
             )
+            # Only restack layers if floating state has changed
+            if needs_change:
+                self.change_layer()
             return
 
         if self._float_state == FloatStates.FULLSCREEN:
+            self._restore_geometry()
             self.floating = False
+            self.change_layer()
             return
 
-    def toggle_fullscreen(self):
-        self.fullscreen = not self.fullscreen
-
     @property
     def maximized(self):
         return self._float_state == FloatStates.MAXIMIZED
@@ -1392,6 +1779,9 @@ class Window(_Window, base.Window):
         if do_maximize:
             screen = self.group.screen or self.qtile.find_closest_screen(self.x, self.y)
 
+            if self._float_state not in (FloatStates.MAXIMIZED, FloatStates.FULLSCREEN):
+                self._save_geometry()
+
             bw = self.group.floating_layout.max_border_width
             self._enablefloating(
                 screen.dx,
@@ -1402,11 +1792,9 @@ class Window(_Window, base.Window):
             )
         else:
             if self._float_state == FloatStates.MAXIMIZED:
+                self._restore_geometry()
                 self.floating = False
 
-    def toggle_maximize(self, state=FloatStates.MAXIMIZED):
-        self.maximized = not self.maximized
-
     @property
     def minimized(self):
         return self._float_state == FloatStates.MINIMIZED
@@ -1420,13 +1808,16 @@ class Window(_Window, base.Window):
             if self._float_state == FloatStates.MINIMIZED:
                 self.floating = False
 
+    @expose_command()
     def toggle_minimize(self):
         self.minimized = not self.minimized
 
-    def cmd_is_visible(self) -> bool:
+    @expose_command()
+    def is_visible(self) -> bool:
         return not self.hidden and not self.minimized
 
-    def cmd_static(
+    @expose_command()
+    def static(
         self,
         screen: int | None = None,
         x: int | None = None,
@@ -1450,7 +1841,7 @@ class Window(_Window, base.Window):
             self.group.remove(self)
         s = Static(self.window, self.qtile, screen, x, y, width, height)
         self.qtile.windows_map[self.window.wid] = s
-        self.qtile.core.update_client_list(self.qtile.windows_map)
+        self.qtile.core.update_client_lists()
         hook.fire("client_managed", s)
 
     def tweak_float(self, x=None, y=None, dx=0, dy=0, w=None, h=None, dw=0, dh=0):
@@ -1485,10 +1876,12 @@ class Window(_Window, base.Window):
 
         self._reconfigure_floating()
 
-    def getsize(self):
+    @expose_command()
+    def get_size(self):
         return (self.width, self.height)
 
-    def getposition(self):
+    @expose_command()
+    def get_position(self):
         return (self.x, self.y)
 
     def _reconfigure_floating(self, new_float_state=FloatStates.FLOATING):
@@ -1504,13 +1897,18 @@ class Window(_Window, base.Window):
                 self.height,
                 self.borderwidth,
                 self.bordercolor,
-                above=True,
+                above=False,
                 respect_hints=True,
             )
         if self._float_state != new_float_state:
             self._float_state = new_float_state
             if self.group:  # may be not, if it's called from hook
                 self.group.mark_floating(self, True)
+            if new_float_state == FloatStates.FLOATING:
+                if self.qtile.config.floats_kept_above:
+                    self.keep_above(enable=True)
+            elif new_float_state == FloatStates.MAXIMIZED:
+                self.move_to_top()
             hook.fire("float_change")
 
     def _enablefloating(
@@ -1527,6 +1925,7 @@ class Window(_Window, base.Window):
         # add to group by position according to _NET_WM_DESKTOP property
         group = None
         index = self.window.get_wm_desktop()
+
         if index is not None and index < len(self.qtile.groups):
             group = self.qtile.groups[index]
         elif index is None:
@@ -1539,6 +1938,7 @@ class Window(_Window, base.Window):
             if group != self.qtile.current_screen.group:
                 self.hide()
 
+    @expose_command()
     def togroup(self, group_name=None, *, switch_group=False, toggle=False):
         """Move window to a specified group
 
@@ -1571,8 +1971,9 @@ class Window(_Window, base.Window):
             self.x += group.screen.x
         group.add(self)
         if switch_group:
-            group.cmd_toscreen(toggle=toggle)
+            group.toscreen(toggle=toggle)
 
+    @expose_command()
     def match(self, match):
         """Match window against given attributes.
 
@@ -1698,7 +2099,9 @@ class Window(_Window, base.Window):
                     self.group.focus(self)
                 elif focus_behavior == "smart":
                     if not self.group.screen:
-                        logger.debug("Ignoring focus request")
+                        logger.debug(
+                            "Ignoring focus request (focus_on_window_activation='smart')"
+                        )
                         return
                     if self.group.screen == self.qtile.current_screen:
                         logger.debug("Focusing window")
@@ -1711,7 +2114,7 @@ class Window(_Window, base.Window):
                     logger.debug("Setting urgent flag for window")
                     self.urgent = True
                 elif focus_behavior == "never":
-                    logger.debug("Ignoring focus request")
+                    logger.debug("Ignoring focus request (focus_on_window_activation='never')")
                 else:
                     logger.debug(
                         "Invalid value for focus_on_window_activation: %s", focus_behavior
@@ -1792,73 +2195,60 @@ class Window(_Window, base.Window):
         elif name == "screen":
             return self.group.screen
 
-    def cmd_kill(self):
-        """Kill this window
-
-        Try to do this politely if the client support
-        this, otherwise be brutal.
-        """
-        self.kill()
-
-    def cmd_move_floating(self, dx, dy):
+    @expose_command()
+    def move_floating(self, dx, dy):
         """Move window by dx and dy"""
         self.tweak_float(dx=dx, dy=dy)
 
-    def cmd_resize_floating(self, dw, dh):
+    @expose_command()
+    def resize_floating(self, dw, dh):
         """Add dw and dh to size of window"""
         self.tweak_float(dw=dw, dh=dh)
 
-    def cmd_set_position_floating(self, x, y):
+    @expose_command()
+    def set_position_floating(self, x, y):
         """Move window to x and y"""
         self.tweak_float(x=x, y=y)
 
-    def cmd_set_size_floating(self, w, h):
+    @expose_command()
+    def set_size_floating(self, w, h):
         """Set window dimensions to w and h"""
         self.tweak_float(w=w, h=h)
 
-    def cmd_place(self, x, y, width, height, borderwidth, bordercolor, above=False, margin=None):
-        self.place(x, y, width, height, borderwidth, bordercolor, above, margin)
-
-    def cmd_get_position(self):
-        return self.getposition()
-
-    def cmd_get_size(self):
-        return self.getsize()
-
-    def cmd_toggle_floating(self):
-        self.toggle_floating()
-
-    def cmd_enable_floating(self):
+    @expose_command()
+    def enable_floating(self):
         self.floating = True
 
-    def cmd_disable_floating(self):
+    @expose_command()
+    def disable_floating(self):
         self.floating = False
 
-    def cmd_toggle_maximize(self):
-        self.toggle_maximize()
+    @expose_command()
+    def toggle_maximize(self):
+        self.maximized = not self.maximized
 
-    def cmd_toggle_fullscreen(self):
-        self.toggle_fullscreen()
+    @expose_command()
+    def toggle_fullscreen(self):
+        self.fullscreen = not self.fullscreen
 
-    def cmd_enable_fullscreen(self):
+    @expose_command()
+    def enable_fullscreen(self):
         self.fullscreen = True
 
-    def cmd_disable_fullscreen(self):
+    @expose_command()
+    def disable_fullscreen(self):
         self.fullscreen = False
 
-    def cmd_toggle_minimize(self):
-        self.toggle_minimize()
-
-    def cmd_bring_to_front(self):
-        if self.floating:
-            self.window.configure(stackmode=StackMode.Above)
-        else:
-            self._reconfigure_floating()  # atomatically above
+    @expose_command()
+    def bring_to_front(self):
+        if self.get_wm_type() != "desktop":
+            self.window.configure(stackmode=xcffib.xproto.StackMode.Above)
 
     def _is_in_window(self, x, y, window):
         return window.edges[0] <= x <= window.edges[2] and window.edges[1] <= y <= window.edges[3]
 
-    def cmd_set_position(self, x, y):
+    @expose_command()
+    def set_position(self, x, y):
         if self.floating:
             self.tweak_float(x, y)
             return
@@ -1867,10 +2257,15 @@ class Window(_Window, base.Window):
             if window == self or window.floating:
                 continue
             if self._is_in_window(curx, cury, window):
-                clients = self.group.layout.clients
-                index1 = clients.index(self)
-                index2 = clients.index(window)
-                clients[index1], clients[index2] = clients[index2], clients[index1]
-                self.group.layout.focused = index2
-                self.group.layout_all()
-                break
+                self.group.layout.swap(self, window)
+                return
+
+    @expose_command
+    def focus(self, warp: bool = True) -> None:
+        """Focus the window."""
+        _Window.focus(self, warp)
+
+        # Focusing a fullscreen window puts it into a different layer
+        # priority group. If it's not there already, we need to move it.
+        if self.fullscreen and not self.previous_layer[4]:
+            self.change_layer()
