@@ -39,10 +39,10 @@ from libqtile.widget import base
 
 
 class _GroupBase(base._TextBox, base.PaddingMixin, base.MarginMixin):
-    defaults = [
+    defaults: list[tuple[str, Any, str]] = [
         ("borderwidth", 3, "Current group border width"),
         ("center_aligned", True, "center-aligned group box"),
-    ]  # type: list[tuple[str, Any, str]]
+    ]
 
     def __init__(self, **config):
         base._TextBox.__init__(self, **config)
@@ -157,7 +157,7 @@ class AGroupBox(_GroupBase):
 
     def _configure(self, qtile, bar):
         _GroupBase._configure(self, qtile, bar)
-        self.add_callbacks({"Button1": partial(self.bar.screen.cmd_next_group, warp=False)})
+        self.add_callbacks({"Button1": partial(self.bar.screen.next_group, warp=False)})
 
     def calculate_length(self):
         return self.box_width(self.qtile.groups) + self.margin_x * 2
@@ -237,6 +237,7 @@ class GroupBox(_GroupBase):
             "Hide groups that have no windows and that are not displayed on any screen.",
         ),
         ("spacing", None, "Spacing between groups" "(if set to None, will be equal to margin_x)"),
+        ("toggle", True, "Enable toggling of group when clicking on same group name"),
     ]
 
     def __init__(self, **config):
@@ -265,20 +266,15 @@ class GroupBox(_GroupBase):
         their label. Groups with an empty string as label are never contained.
         Groups that are not named in visible_groups are not returned.
         """
+        groups = filter(lambda g: g.label, self.qtile.groups)
+
         if self.hide_unused:
-            if self.visible_groups:
-                return [
-                    g
-                    for g in self.qtile.groups
-                    if g.label and (g.windows or g.screen) and g.name in self.visible_groups
-                ]
-            else:
-                return [g for g in self.qtile.groups if g.label and (g.windows or g.screen)]
-        else:
-            if self.visible_groups:
-                return [g for g in self.qtile.groups if g.label and g.name in self.visible_groups]
-            else:
-                return [g for g in self.qtile.groups if g.label]
+            groups = filter(lambda g: g.windows or g.screen, groups)
+
+        if self.visible_groups:
+            groups = filter(lambda g: g.name in self.visible_groups, groups)
+
+        return list(groups)
 
     def get_clicked_group(self):
         group = None
@@ -325,7 +321,7 @@ class GroupBox(_GroupBase):
 
     def go_to_group(self, group):
         if group:
-            if self.bar.screen.group != group or not self.disable_drag:
+            if self.bar.screen.group != group or not self.disable_drag or not self.toggle:
                 self.bar.screen.set_group(group, warp=False)
             else:
                 self.bar.screen.toggle_group(group, warp=False)
@@ -335,7 +331,7 @@ class GroupBox(_GroupBase):
         if button not in (5, 4):
             group = self.get_clicked_group()
             if group and self.clicked:
-                group.cmd_switch_groups(self.clicked.name)
+                group.switch_groups(self.clicked.name)
                 self.clicked = None
 
     def calculate_length(self):
@@ -345,7 +341,7 @@ class GroupBox(_GroupBase):
         return width
 
     def group_has_urgent(self, group):
-        return len([w for w in group.windows if w.urgent]) > 0
+        return any(w.urgent for w in group.windows)
 
     def draw(self):
         self.drawer.clear(self.background or self.bar.background)
